-----------------------------------------------------------------------------------------------
-- Client Lua Script for ZinMarker
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "Unit"
require "GameLib"
require "Apollo"
 
-----------------------------------------------------------------------------------------------
-- ZinMarker Module Definition
-----------------------------------------------------------------------------------------------
local ZinMarker = {}
local log

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
local markerEnum = {
	["Bomb"] = 1,
	["Chicken"] = 2,
	["Ghost"] = 3,
	["Mask"] = 4,
	["Octopus"] = 5,
	["Pig"] = 6,
	["Toaster"] = 7,
	["UFO"] = 8,
	["Number_1"] = 9,
	["Number_2"] = 11,
	["Number_3"] = 12,
	["Number_4"] = 13,
	["Number_5"] = 14,
	["Number_5"] = 15,
	["Number_7"] = 16,
	["Number_8"] = 17,
	["Number_9"] = 18,
	["Number_0"] = 19
}

local activeMarkers = {}
local markers = {
	-- name(string), sprite(string), set(bool), location(table), options(table)
	Bomb = {	
		name = "Bomb", 
		sprite = "Icon_Windows_UI_CRB_Marker_Bomb", 
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Chicken = {
		name = "Chicken", 
		sprite = "Icon_Windows_UI_CRB_Marker_Chicken",
		pixieId = nil, 
		location = nil, 
		options = { w = 20, h = 20}
	},
	Ghost = {	
		name = "Ghost", 
		sprite = "Icon_Windows_UI_CRB_Marker_Ghost",
		pixieId = nil, 
		location = nil, 
		options = { w = 20, h = 20}
	},
	Mask = {	
		name = "Mask", 
		sprite = "Icon_Windows_UI_CRB_Marker_Mask",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Octopus = {
		name = "Octopus", 
		sprite = "Icon_Windows_UI_CRB_Marker_Octopus",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Pig = {	
		name = "Pig", 
		sprite = "Icon_Windows_UI_CRB_Marker_Pig",
		pixieId = nil, 
		location = nil, 
		options = { w = 20, h = 20}
	},
	Toaster = {
		name = "Toaster", 
		sprite = "Icon_Windows_UI_CRB_Marker_Toaster",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	UFO = {		
		name = "UFO", 
		sprite = "Icon_Windows_UI_CRB_Marker_UFO",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_1 = {		
		name = "Number_1", 
		sprite = "sprFloater_Critical1",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_2 = {		
		name = "Number_2", 
		sprite = "sprFloater_Critical2",
		pixieId = nil, 
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_3 = {		
		name = "Number_3", 
		sprite = "sprFloater_Critical3",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_4 = {		
		name = "Number_4",
		sprite = "sprFloater_Critical4",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_5 = {		
		name = "Number_5", 
		sprite = "sprFloater_Critical5",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_6 = {		
		name = "Number_6", 
		sprite = "sprFloater_Critical6",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_7 = {		
		name = "Number_7", 
		sprite = "sprFloater_Critical7",
		pixieId = nil, 
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_8 = {		
		name = "Number_8", 
		sprite = "sprFloater_Critical8",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_9 = {		
		name = "Number_9", 
		sprite = "sprFloater_Critical9",
		pixieId = nil, 
		location = nil, 
		options = { w = 20, h = 20}
	},
	Number_0 = {		
		name = "Number_0", 
		sprite = "sprFloater_Critical0",
		pixieId = nil,
		location = nil, 
		options = { w = 20, h = 20}
	}
}
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function ZinMarker:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	self.setMarker = false
	self.markerToSet = nil
	self.windowVisible = true
	self.optionWindowVisible = false
	self.savedProfiles = {}
	self.options = {
		title = "ZinMarker",
		buttons = {
			w = 34,
			h = 34
		},
		shareMarker = true,
		window = {
			main = {
				l = 0,
				t = 0,
				r = 581,
				b = 51
			},
			option = {
				l = 10,
				t = 80,
				r = 470,
				b = 490
			}
		}
	}
    return o
end

function ZinMarker:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		"Gemini:Logging-1.2"
	}

    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-----------------------------------------------------------------------------------------------
-- ZinMarker OnLoad
-----------------------------------------------------------------------------------------------
function ZinMarker:OnLoad()
	local GeminiLogging = Apollo.GetPackage("Gemini:Logging-1.2").tPackage
	log = GeminiLogging:GetLogger({
 		level = GeminiLogging.INFO,
  		pattern = "%l - %m",
  		appender = "GeminiConsole"
	})
	
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("ZinMarker.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	self.shareChannel = ICCommLib.JoinChannel("ZinMarker", "OnReceiveMarker", self)
end

-----------------------------------------------------------------------------------------------
-- ZinMarker OnDocLoaded
-----------------------------------------------------------------------------------------------
function ZinMarker:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = 		Apollo.LoadForm(self.xmlDoc, "ZinMarker", nil, self)
		self.wndOptions = 	Apollo.LoadForm(self.xmlDoc, "ZinMarkerOptions", nil, self)
		self.overlay = 		Apollo.LoadForm(self.xmlDoc, "Overlay", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
		-- Register handlers for events, slash commands and timer, etc.
		Apollo.RegisterSlashCommand("zm", "OnZinMarkerOn", self)
		Apollo.RegisterEventHandler("VarChange_FrameCount", "OnNextFrame", self)		

		-- Do additional Addon initialization here
		self.wndMain:SetAnchorOffsets(
			self.options.window.main.l,
			self.options.window.main.t,
			self.options.window.main.r,
			self.options.window.main.b)
		

		self.wndOptions:SetAnchorOffsets(
			self.options.window.option.l,
			self.options.window.option.t,
			self.options.window.option.r,
			self.options.window.option.b)
			
		self.wndMain:Show(self.windowVisible, true)
		self.wndOptions:Show(self.optionWindowVisible, true)
	end
end

function ZinMarker:SetSpritesForButtons()
	sprite = {}
	sprite.w = 34
	sprite.h = 34
	
	for i = 1, self.markers do
		local marker = self.markers[i]
		lof:info("Set marker " .. marker[0])
	end
end

-----------------------------------------------------------------------------------------------
-- ZinMarker Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

-- on SlashCommand "/zm"
function ZinMarker:OnZinMarkerOn()
	if self.windowVisible then
		self.windowVisible = false 
	else
		self.windowVisible = true
	end 
	self.wndMain:Show(self.windowVisible , true)
end

-- on timer
function ZinMarker:OnNextFrame()
	self:DrawMarkers()
end

function ZinMarker:DrawMarkers()
	for key, marker in pairs(activeMarkers) do
		local pixie = self:GenerateMarkerPixie(marker)
		self.overlay:UpdatePixie(marker.pixieId, pixie)
	end
end

-----------------------------------------------------------------------------------------------
-- ZinMarker Marker Functions
-----------------------------------------------------------------------------------------------
function ZinMarker:SetMarker(marker, loc)
	-- clear old position
	local index = self:GetMarkerIndex(marker)
	if index ~= nil then	
		self:ClearMarker(activeMarkers[index])
		table.remove(activeMarkers, index)
	end
	
	-- set new marker
	marker.location = loc
	local pixie = self:GenerateMarkerPixie(marker)
	marker.pixieId = self.overlay:AddPixie(pixie) -- safe pixie id for position updates
	table.insert(activeMarkers, marker)
	
	self.markerToSet = nil
	self.setMarkers = false
end

function ZinMarker:ClearMarker(marker)
	if marker.pixieId ~= nil then
		self.overlay:DestroyPixie(marker.pixieId)
		marker.pixieId = nil
		marker.location = nil
	end
end

-----------------------------------------------------------------------------------------------
-- ZinMarkerForm Communication Functions
-----------------------------------------------------------------------------------------------
function ZinMarker:ShareMarker(marker)
	if GroupLib.GetMemberCount() > 0 and self.options.shareMarker then
		local msg = {}
		msg.Destination = ""
		msg.Sender = GameLib.GetPlayerUnit():GetName()
		msg.DestinationType = "group"
		msg.Marker = self:MarkerToString(marker)
		self.shareChannel:SendMessage(msg)
	end
end

function ZinMarker:OnReceiveMarker(chan, msg)
	if msg.DestinationType == "group" and GroupLib.GetMemberCount() > 0 then
		for i = 1, GroupLib.GetMemberCount() do
			if GroupLib.GetGroupMember(i).strCharacterName == msg.Sender then
				local markerInfo =	self:GetMarkerInfoFromString(msg.Marker)
				self:SetMarker(markerInfo, markerInfo.loc)
			end
		end
	end
end

-----------------------------------------------------------------------------------------------
-- ZinMarkerForm Helper Functions
-----------------------------------------------------------------------------------------------
function ZinMarker:GetMarkerIndex(marker)
	local index = 0
	for i = 1, table.getn(activeMarkers) do
		if activeMarkers[i].name == marker.name then
			return i
		end
	end
	return nil
end

-- copied from WorldMarkers
function ZinMarker:GenerateMarkerPixie(marker)
	local loc =  GameLib.WorldLocToScreenPoint(marker.location)
	local w = self.options.buttons.w /2
	local h = self.options.buttons.h /2
	return {
		strSprite = marker.sprite,
		loc = {
			nOffsets = {
				loc.x-(w),
				loc.y-(h),
				loc.x+(w),
				loc.y+(h)
			}
		}
	}
end

function ZinMarker:MarkerToString(marker)
	return marker.name .. ";" .. marker.location.x .. ";" 
			.. marker.location.y .. ";" .. marker.location.z 
			.. ";" .. marker.sprite
end

function ZinMarker:GetMarkerInfoFromString(string)
	local marker = {}
	local counter = 1
	for coord in string.gmatch(string, '([^;]+)') do
    	marker[counter] = coord
		counter = counter + 1
	end	
					
	return {
		loc = Vector3.New(marker[2],marker[3],marker[4]),
		name = marker[1],
		sprite = marker[5]
	}
end

-----------------------------------------------------------------------------------------------
-- ZinMarker Options Window
-----------------------------------------------------------------------------------------------
function ZinMarker:ShowOptions(wndHandler, wndControl, eMouseButton)
	if self.optionWindowVisible then
		self.optionWindowVisible = false
	else
		self.optionWindowVisible = true
		self:LoadMarkerOptions(wndHandler, self.wndOptions:FindChild("Marker"), eMouseButton)
	end
	self.wndOptions:Show(self.optionWindowVisible , true)
end

function ZinMarker:LoadMarkerOptions(wndHandler, wndControl, eMouseButton)
	local optionStr = "Marker"
	self.wndOptions:FindChild("Options"):FindChild("OptionsTitle"):SetText(optionStr)
	local container = self.wndOptions:FindChild("Options"):FindChild("OptionContainer")
	container:DestroyChildren()
	local form = Apollo.LoadForm(self.xmlDoc, "MarkerOptions", container, self)
	form:FindChild("MarkerSize"):SetText(self.options.buttons.w)
	form:FindChild("Share"):SetCheck(self.options.shareMarker)
end

function ZinMarker:OnChangeMarkerSize(wndHandler, wndControl, strText)
	local value = tonumber(strText)
	if value ~= nil then
		self.options.buttons.w = value
		self.options.buttons.h = value
	end
end

function ZinMarker:OnChangeShareMarker(wndHandler, wndControl, eMouseButton)
	if wndControl:IsChecked() then
		self.options.shareMarker = true
	else
		self.options.shareMarker = true
	end
end

function ZinMarker:LoadProfileOptions()
	local optionStr = "Profile"
	self.wndOptions:FindChild("Options"):FindChild("OptionsTitle"):SetText(optionStr)
	local container = self.wndOptions:FindChild("Options"):FindChild("OptionContainer")
	container:DestroyChildren()
	Apollo.LoadForm(self.xmlDoc, "ProfileSave", container, self)
	self:AddProfileList(container)
end

function ZinMarker:AddProfileList(container)
	for _, profile in pairs(self.savedProfiles) do
		local selectProfile = Apollo.LoadForm(self.xmlDoc, "Profile", container, self)
		selectProfile:FindChild("LoadProfile"):SetText(profile.name)
	end
	container:ArrangeChildrenVert()
end

function ZinMarker:SaveCurrentMarkers(wndHandler, wndControl, eMouseButton)
	local profileName  = wndControl:GetParent():FindChild("ProfileName"):GetText()
	local somethingToSave = false
	-- profile name has to contain at least one charachter
	if profileName ~= nil and profileName ~= "" then
		local markersToSave = {}
		for _,marker in pairs(activeMarkers) do
			table.insert(markersToSave, self:MarkerToString(marker))
		end	
		if table.getn(markersToSave) > 0 then
			local profile = {
					name = profileName,
					markers = markersToSave
			}
			local index = self:GetProfileIndex(profileName)
			if index ~= nil then
				self.savedProfiles[index] = profile 
			else
				table.insert(self.savedProfiles, profile)
			end
			if self.optionWindowVisible then
				self:LoadProfileOptions()
			end
		end
	end
end

function ZinMarker:DeleteProfile(wndHandler, wndControl, eMouseButton)
	local profileName  = wndControl:GetParent():FindChild("LoadProfile"):GetText()
	local index = self:GetProfileIndex(profileName)
	local counter = 1
	table.remove(self.savedProfiles, index)
	self:LoadProfileOptions()
end

function ZinMarker:LoadProfile(wndHandler, wndControl, eMouseButton)
	local profileName  = wndControl:GetText()
	local profile
	for _, temp in pairs(self.savedProfiles) do
		if temp.name == profileName then
			profile = temp
		end
	end
	self:ResetAllMarker()
	for key, markerString in pairs(profile.markers) do
		local markerInfo = self:GetMarkerInfoFromString(markerString)
		local marker = {
			name = markerInfo.name,
			sprite = markerInfo.sprite,
			loc = markerInfo.loc
		}
		
		self:SetMarker(marker, markerInfo.loc)
		
		self:ShareMarker(marker)
	end
end

function ZinMarker:GetProfileIndex(profileName)
	local index = nil
	local counter = 1
	for _, temp in pairs(self.savedProfiles) do
		if temp.name == profileName then
			index = counter
		end
		counter = counter + 1
	end	
	return index
end

-----------------------------------------------------------------------------------------------
-- ZinMarker Bar functions
-----------------------------------------------------------------------------------------------
function ZinMarker:ActivateMarker(wndHandler, wndControl, eMouseButton)
	self.setMarker = true
	self.markerToSet = wndControl:GetName()
	
	local marker = markers[self.markerToSet]
	local pos = GameLib.GetPlayerUnit():GetPosition()
	local vec = Vector3.New(pos.x, pos.y, pos.z)
	self:SetMarker(marker, vec)
	
	self:ShareMarker(marker)
end

function ZinMarker:ResetMarker(wndHandler, wndControl, eMouseButton)
	local btnName = wndControl:GetName()
	local markerStr  = string.gsub(wndControl:GetName(), "Reset", "")
	local marker = markers[markerStr  ]
	self:ClearMarker(marker)
	table.remove(activeMarkers, self:GetMarkerIndex(marker)
)
end

function ZinMarker:ResetAllMarker()
	for key, marker in pairs(activeMarkers) do
		self:ClearMarker(marker)
	end
	activeMarkers = {}
end

-----------------------------------------------------------------------------------------------
-- ZinMarker Event Handler
-----------------------------------------------------------------------------------------------
function ZinMarker:OnSave(eLevel)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return
    end
	local saveData = {
		profiles = self.savedProfiles,
		wndMainPos = {},
		options = self.options,
	}
	saveData.options.window.main.l, saveData.options.window.main.t, saveData.options.window.main.r, saveData.options.window.main.b = self.wndMain:GetAnchorOffsets()
	saveData.options.window.option.l, saveData.options.window.option.t, saveData.options.window.option.r, saveData.options.window.option.b = self.wndOptions:GetAnchorOffsets()
	return saveData
end

function ZinMarker:OnRestore(eLevel, tData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end
	if tData then
		if tData.profiles then
			self.savedProfiles = tData.profiles
		end
		if tData.options then
			self.options = tData.options
		end
	end
end

-----------------------------------------------------------------------------------------------
-- ZinMarker Instance
-----------------------------------------------------------------------------------------------
local ZinMarkerInst = ZinMarker:new()
ZinMarkerInst:Init()

-------------------------------------------------------------------------------------
-- maybe later if you can get mouse click coordinates in world
function ZinMarker:OnGameClickWorld(loc)
	if setMarker then	
		self:SetMarker(markers[markerToSet], loc)
	end
end
